/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file bspline.h

    \brief defines class bspline

  class bspline is the most convenient access to vspline's functionality.
  It attempts to do 'the right thing' by automatically creating suitable helper
  objects and parametrization so that the spline does what it's supposed to do.
  Most users will not need anything else, and using class bspline is quite
  straightforward. It's quite possible to have a b-spline up and running with
  a few lines of code without even having to make choices concerning it's
  parametrization, since there are sensible defaults for everything. At the same
  time, pretty much everything *can* be parametrized even at this level.
  bspline objects can be used without any knowledge of their internals,
  e.g. as parameters to the remap functions.
  
  Note that class bspline does not provide evaluation of the spline. To evaluate,
  objects of class evaluator (see eval.h) are used, which construct from a bspline
  object with additional parameters, like, whether to calculate the spline's
  value or it's derivative(s) or whether to use optimizations for special cases.
  
  While using 'raw' coefficient arrays with an evaluation scheme which applies
  boundary conditions is feasible and most memory-efficient, it's not so well
  suited for very fast evaluation, since the boundary treatment needs conditionals,
  and 'breaks' uniform access, which is especially detrimental when using
  vectorization. So vspline uses coefficient arrays with a few extra coefficients
  'framing' or 'bracing' the 'core' coefficients. Since evaluation of the spline
  looks at a certain small section of coefficients (the evaluator's 'support'),
  the bracing is chosen so that this lookup will always succeed without having to
  consider boundary conditions: the brace is set up to make the boundary conditions
  explicit, and the evaluation can proceed blindly without bounds checking. With
  large coefficient arrays, the little extra space needed for the brace becomes
  negligible, but the code for evaluation becomes much faster and simpler.

  So class bspline handles several views to the coefficients it operates on, these
  are realized as vigra::MultiArrayViews, and they all share the same storage:

  - the 'core', which is a view to an array of data precisely the same shape as
    the knot point data over which the spline is calculated.

  - 'coeffs', which is a view to the core, plus 'bracing' needed to evaluate
    the spline with vspline's evaluation code. 'coeffs' contains 'core'.
    I refer to this view as the 'braced coefficients' as well.

  - 'container', which contains the two views above plus an additional frame
    of coefficients used for the 'explicit' scheme of extrapolation before
    prefiltering, or as extra 'headroom' if 'shifting' the spline is intended. 

  Using class bspline, there is a choice of 'strategy'. The simplest strategy is
  'UNBRACED'. With this strategy, after putting the knot point data into the bspline's
  'core' area and calling prefilter(), the core area will contain the b-spline
  coefficients. The resulting b-spline object can't be evaluated with the code in eval.h.
  this mode of operation is intended for users who want to do their own processing of the
  coefficients and don't need the code in eval.h. prefiltering is done using an implicit
  scheme as far as the boundary conditions are concerned.
  
  The 'standard' strategy is 'BRACED'. Here, after prefiltering, the view 'coeffs'
  in the bspline object will contain the b-spline coefficients, surrounded by a 'brace'
  of coefficients which allows code in eval.h to process them without special treatment
  for values near the border (the brace covers what support is needed by marginal
  coefficients). Again, an implicit scheme is used.
  
  The third strategy, 'EXPLICIT', extrapolates the knot point data in the 'core' area
  sufficiently far to suppress margin effects when the prefiltering is performed without
  initial coefficient calculation. If the 'frame' of extrapolated data is large enough,
  the result is just the same. The inner part of the frame is taken as the brace, so no
  bracing needs to be performed explicitly. The resulting b-spline object will work with
  vspline's evaluation code. Note that the explicit scheme uses 'GUESS' boundary conditions
  on the (framed) array, which tries to minimize margin effects further.

  Also note that the additional memory needed for the 'frame' will be held throughout the bspline
  object's life, the only way to 'shrink' the coefficient array to the size of the braced or core
  coefficients is by copying them out to a smaller array.

  The fourth strategy, 'MANUAL', is identical to 'EXPLICIT', except that automatic extrapolation
  of the core data to the frame is not performed. Instead, this strategy relies on the user to
  fill the frame with extrapolated data. This is to allow for the user to apply custom
  extrapolation schemes. The procedure would be to create the bspline object, fill the core,
  apply the extrapolation, then call prefilter.

  Probably the most common scenario is that the source data for the spline are available from
  someplace like a file. Instead of reading the file's contents into memory first and passing
  the memory to class bspline, there is a more efficient way: a bspline object is set up
  first, with the specification of the size of the incoming data and the intended mode of
  operation. The bspline object allocates the memory it will need for the purpose, but
  doesn't do anything else. The 'empty' bspline object is then 'filled' by the user
  by putting data into it's 'core' area. Subsequently, prefilter() is called, which converts
  the data to b-spline coefficients. This way, only one block of memory is used throughout,
  the initial data are overwritten by the coefficients, operation is in-place and most efficient.

  If this pattern can't be followed, there are alternatives:

  - if a view to an array at least the size of the container array is passed into bspline's
    constructor, this view is 'adopted' and all operations will use the data it refers to.
    The caller is responsible for keeping these data alive while the bspline object exists,
    and relinquishes control over the data, which may be changed by the bspline object.
    Note that there is a convenience method, 'container_size', which can calculate the
    shape of a container suitable for the purpose.

  - if data are passed to prefilter(), they will be taken as containing the knot point data,
    rather than expecting the knot point data to be in the bspline oject's memory already.
    This can also be used to reuse a bspline object with new data. The data passed in will
    not be modified. This is most efficient when using an implicit scheme; when used together
    with EXPLICIT, the data are (automatically) copied into the core area before prefiltering,
    which is unnecessary with the implicit schemes - they can 'pull in' data in the course
    of their operation.

  While there is no explicit code to create a 'smoothing spline' - a b-spline evaluating
  the source data without prefiltering them - this can be achieved simply by creating a b-spline
  object with spline degree 0 and 'shifting' it to the desired degree for evaluation. Note that
  you'll need the EXPLICIT strategy for the purpose, or specify extra 'headroom', because
  otherwise the spline won't have enough 'headroom' for shifting.
  
  If stronger smoothing is needed, this can be achieved with the code in filter.h, passing in
  appropriate pole values. a single-pole filter with a positive pole in ] 0 , 1 [ will do the
  trick - the larger the pole, the stronger the smoothing. Note that smoothing with large pole
  values will need a large 'horizon' as well to handle the margins properly.

  With shifting, you can also create a 'poor man's pyramid'. While using no additional storage,
  you can extract smoothed data from the spline by shifting it up. This only goes so far, though,
  because even a degree-20 b-spline reconstruction kernel's equivalent gaussian doesn't have a very
  large standard deviation, and evaluation times become very long. From the gaussian approximation
  of the b-spline basis function, it can be seen that the equivalent gaussian's standard deviation is
  ( degree + 1 ) / 12.0, so a quintic spline will have a standard deviation of 0.5 only.
*/

#ifndef VSPLINE_BSPLINE_H
#define VSPLINE_BSPLINE_H

#include "prefilter.h"
#include "brace.h"

namespace vspline {

/// This enumeration is used to determine the prefiltering scheme to be used.

typedef enum { UNBRACED , ///< implicit scheme, no bracing applied
               BRACED ,   ///< implicit scheme, bracing will be applied
               EXPLICIT , ///< explicit scheme, frame with extrapolated signal, brace
               MANUAL     ///< like explicit, but don't frame before filtering
} prefilter_strategy  ;

/// struct bspline is a convenience class which bundles a coefficient array (and it's creation)
/// with a set of metadata describing the parameters used to create the coefficients and the
/// resulting data. I have chosen to implement class bspline so that there is only a minimal
/// set of template arguments, namely the spline's data type (like pixels etc.) and it's dimension.
/// All other parameters relevant to the spline's creation are passed in at construction time.
/// This way, if explicit specialization becomes necessary (like, to interface to code which
/// can't use templates) the number of specializations remains manageable. This design decision
/// pertains specifically to the spline's degree, which can also be implemented as a template
/// argument, allowing for some optimization by making some members static. Yet going down this
/// path requires explicit specialization for every spline degree used and the performance gain
/// I found doing so was hardly measurable, while automatic testing became difficult and compilation
/// times grew.
///
/// I chose making bspline a struct for now, but messing with the data inside is probably
/// not a good idea...

template < class value_type , int _dimension >
struct bspline
{
  /// pull the template arg into an enum
  enum { dimension = _dimension } ;
  /// if the coefficients are owned, this array holds the data
  typedef vigra::MultiArray < dimension , value_type > array_type ;
  /// data are read and written to vigra MultiArrayViews
  typedef vigra::MultiArrayView < dimension , value_type > view_type ;
  /// multidimensional index type
  typedef typename view_type::difference_type shape_type ;
  /// nD type for one boundary condition per axis
  typedef typename vigra::TinyVector < bc_code , dimension > bcv_type ;

  /// elementary type of value_type, like float or double
  typedef typename ExpandElementResult < value_type >::type real_type ;
  enum { channels = ExpandElementResult < value_type >::size } ;

  typedef bspline < real_type , dimension > channel_view_type ;
  
  // for internal calculations in the filter, we use the elementary type of value_type.
  // Note how in class bspline we make very specific choices about the
  // source data type, the target data type and the type used for arithmetics: we use
  // the same value_type for source and target array and perform the arithmetics with
  // this value_type's elementary type. The filter itself is much more flexible; all of
  // the three types can be different, the only requirements are that the value_types
  // must be vigra element-expandable types with an elementary type that can be cast
  // to and from math_type, and math_type must be a real data type, with the additional
  // requirement that it can be vectorized by VC if Vc is used.

  typedef real_type math_type ; // arbitrary, can use float or double here.

private:

  array_type _coeffs ;
  prefilter_strategy strategy ;

public:

  const bcv_type bcv ;      ///< coundary conditions, see common.h
  
  view_type container ;     ///< view to container array
  view_type coeffs ;        ///< view to the braced coefficient array
  view_type core ;          ///< view to the core part of the coefficient array
  int spline_degree ;       ///< degree of the spline (3 == cubic spline)
  double tolerance ;        ///< acceptable error
  double smoothing ;        ///< E ] 0 : 1 [; apply smoothing to data before prefiltering
  bool braced ;             ///< whether coefficient array is 'braced' or not
  int horizon ;             ///< additional frame size for explicit scheme
  shape_type left_brace ;   ///< width(s) of the left handside bracing
  shape_type right_brace ;  ///< width(s) of the right handside bracing
  shape_type left_frame ;   ///< width(s) of the left handside bracing
  shape_type right_frame ;  ///< width(s) of the right handside bracing
  shape_type container_shape ;   ///< shape of the container array
  shape_type core_shape ;   ///< shape of the core coefficient array
  shape_type braced_shape ; ///< shape of the coefficient array + bracing

  /// lower_limit returns the lower bound of the spline's defined range.
  /// This is usually 0.0, but with REFLECT boundary condition it's -0.5,
  /// the lower point of reflection. The lowest coordinate at which the
  /// spline can be accessed may be lower: even splines have wider support,
  /// and splines with extra headroom add even more room to manoevre.

  double lower_limit ( const int & axis )
  {
    double limit = 0.0 ;
    
    if ( bcv [ axis ] == vspline::REFLECT )
      limit = -0.5 ;
    
    return limit ;
  }
  
  /// upper_limit returns the upper bound of the spline's defined range.
  /// This is normally M - 1 if the shape for this axis is M. Splines with
  /// REFLECT boundary condition use M - 0.5, the upper point of reflection,
  /// and periodic splines use M. The highest coordinate at which the spline
  /// may be accessed safely may be higher.
  
  double upper_limit ( const int & axis )
  {
    double limit = core.shape [ axis ] - 1 ;
    
    if ( bcv [ axis ] == vspline::REFLECT )
      limit += 0.5 ;
    else if ( bcv [ axis ] == vspline::PERIODIC )
      limit += 1.0 ;
    
    return limit ;
  }
  
  /// setup_metrics determines the sizes of the three views and any braces/frames
  /// needed with the given parameters

  void setup_metrics ( int headroom = 0 )
  {
    switch ( strategy )
    {
      case UNBRACED:
        // UNBRACED is simple: all internal views are the same. prefiltering will
        // be done using an implicit scheme.
        container_shape = braced_shape = core_shape ;
        left_brace = right_brace = left_frame = right_frame = shape_type() ;
        braced = false ;
        return ;
      case BRACED:
        // again an implicit prefiltering scheme will be used, but here we add
        // a 'brace' to the core data, which makes the resulting bspline object
        // suitable to work with vspline's evaluation code. The container array's
        // size is the same as the braced core's size, unless additional headroom
        // was requested.
        braced_shape = bracer<view_type>::target_shape ( core_shape , bcv , spline_degree ) ;
        left_brace = bracer<view_type>::left_corner ( bcv , spline_degree ) ;
        right_brace = bracer<view_type>::right_corner ( bcv , spline_degree ) ;
        left_frame = left_brace + shape_type(headroom) ;
        right_frame = right_brace + shape_type(headroom) ;
        break ;
      case EXPLICIT:
        // here we prepare for an explicit extrapolation. This requires additional
        // space, namely the 'frame', around the core data, into which the extrapolated
        // data are put before prefiltering the lot. This frame is applied in excess of
        // the bracing, to make sure all coefficients inside the brace meet the precision
        // requirements expressed by the choice of 'horizon'. If additional headroom
        // is requested, this comes yet on top.
        braced_shape = bracer<view_type>::target_shape ( core_shape , bcv , spline_degree ) ;
        left_brace = bracer<view_type>::left_corner ( bcv , spline_degree ) ;
        right_brace = bracer<view_type>::right_corner ( bcv , spline_degree ) ;
        left_frame = left_brace + shape_type(horizon + headroom) ;
        right_frame = right_brace + shape_type(horizon + headroom) ;
        break ;
      case MANUAL:
        // nothing to do here, it's up to the user
        break ;
    }
    braced = true ;
    
    // for odd splines with REFLECT boundary conditions we increase the size of
    // the left frame, so that access to coordinates in [-0.5:0] will not result
    // in accessing memory outside the coefficient array. Access to [M-1:M-0.5]
    // is safe with the right brace size the bracer has returned.

    if ( spline_degree & 1 )
    {
      for ( int d = 0 ; d < dimension ; d++ )
      {
        if ( bcv[d] == REFLECT || bcv[d] == SPHERICAL )
          left_frame[d]++ ;
      }
    }

    container_shape = core_shape + left_frame + right_frame ;
  }

  /// this method calculates the size of container needed by a bspline object with
  /// the given parameters. This is a helper routine for use cases where the memory for
  /// the bspline object is allocated externally and passed into the bspline object.

  static shape_type container_size ( shape_type core_shape ,  ///< shape of knot point data
            int spline_degree = 3 ,                ///< spline degree with reasonable default
            bcv_type bcv = bcv_type ( MIRROR ) ,   ///< boundary conditions and common default
            prefilter_strategy strategy = BRACED , ///< default strategy is the 'implicit' scheme
            int horizon = sizeof(real_type) * 3 )  ///< width of frame for explicit scheme (heuristic)
  {
    switch ( strategy )
    {
      case UNBRACED:
      {
        return core_shape ;
        break ;
      }
      case BRACED:
      {
        auto ts = bracer<view_type>::target_shape
          ( core_shape , bcv , spline_degree ) ;

        if ( spline_degree & 1 )
        {
          for ( int d = 0 ; d < dimension ; d++ )
          {
            if ( bcv[d] == REFLECT || bcv[d] == SPHERICAL )
              ts[d]++ ;
          }
        }
        return ts ;
        break ;
      }
      case EXPLICIT:
      {
        auto ts = bracer<view_type>::target_shape
          ( core_shape , bcv , spline_degree ) ;

        if ( spline_degree & 1 )
        {
          for ( int d = 0 ; d < dimension ; d++ )
          {
            if ( bcv[d] == REFLECT || bcv[d] == SPHERICAL )
              ts[d]++ ;
          }
        }
        ts += 2 * horizon ;
        return ts ;
        break ;
      }
      case MANUAL:
        // nothing to do here, it's up to the user
        break ;
    }
    return core_shape ;
  }

  /// construct a bspline object with appropriate storage space to contain and process an array
  /// of knot point data with shape core_shape. Depending on the strategy chosen and the other
  /// parameters passed, more space than core_shape may be allocated. Once the bspline object
  /// is ready, usually it is filled with the knot point data and then the prefiltering needs
  /// to be done. This sequence assures that the knot point data are present in memory only once,
  /// the prefiltering is done in-place. So the user can create the bspline, fill in data (like,
  /// from a file), prefilter, and then evaluate.
  ///
  /// alternatively, if the knot point data are already manifest elsewhere, they can be passed
  /// to prefilter(). With this mode of operation, they are 'pulled in' during prefiltering.
  ///
  /// It's possible to pass in a view to an array providing space for the coefficients,
  /// or even the coefficients themselves. This is done via the parameter _space. This has
  /// to be an array of the same or larger shape than the container array would end up having
  /// given all the other parameters. This view is then 'adopted' and subsequent processing
  /// will operate on it's data. container_size can be used to get the precise shape of the memory
  /// needed with the given parameters.
  ///
  /// with the EXPLICIT scheme, the horizon is set by default to a value which is
  /// deemed to be 'sufficiently large' to keep the error 'low enough'. the expression
  /// used here produces a frame which is roughly the size needed to make any margin
  /// effects vanish by the time the prefilter hits the core, but it's a bit 'rule of thumb'.
  ///
  /// The additional parameter 'headroom' is used to make the 'frame' even wider. This is
  /// needed if the spline is to be 'shifted' up (evaluated as if it had been prefiltered
  /// with a higher-degree prefilter) - see shift().
  ///
  /// While bspline objects allow very specific parametrization, most use cases won't use
  /// parameters beyond the first few. The only mandatory parameter is, obviously, the
  /// shape of the knot point data, the original data which the spline is built over.
  /// This shape 'returns' as the bspline object's 'core' shape. If this is the only
  /// parameter passed to the constructor, the resulting bspline object will be a
  /// cubic b-spline with mirror boundary conditions, generated with an implicit
  /// extrapolation scheme to a 'good' quality, no smoothing, and allocating it's own
  /// storage for the coefficients, and the resuling bspline object will be suitable for
  /// use with vspline's evaluation code.
  
  // TODO: when bracing/framing is applied, we might widen the array size to a
  // multiple of the Vc:Vector's Size for the given data type to have better-aligned
  // access. This may or may not help, has to be tested. We might also want to position
  // the origin of the brace to an aligned position, since evaluation is based there.
  
  // TODO: while the coice to keep the value_types and math_type closely related makes
  // for simple code, with the more flexible formulation of the prefiltering code we might
  // widen class bspline's scope to accept input of other types and/or use a different
  // math_type.

  bspline ( shape_type _core_shape ,                ///< shape of knot point data
            int _spline_degree = 3 ,                ///< spline degree with reasonable default
            bcv_type _bcv = bcv_type ( MIRROR ) ,   ///< boundary conditions and common default
            prefilter_strategy _strategy = BRACED , ///< default strategy is the 'implicit' scheme
            int _horizon = -1 ,                     ///< width of frame for explicit scheme
            double _tolerance = -1.0 ,              ///< acceptable error (relative to unit pulse)
            double _smoothing = 0.0 ,               ///< apply smoothing to data before prefiltering
            int headroom = 0 ,                      ///< additional headroom, for 'shifting'
            view_type _space = view_type()          ///< coefficient storage to 'adopt'
          )
  : core_shape ( _core_shape ) ,
    spline_degree ( _spline_degree ) ,
    bcv ( _bcv ) ,
    smoothing ( _smoothing ) ,
    strategy ( _strategy )
  {
    if ( _tolerance < 0.0 )
    {
      // heuristic: 'reasonable' defaults
      if ( std::is_same < real_type , float > :: value )
        tolerance = .000001 ;
      else if ( std::is_same < real_type , double > :: value )
        tolerance = .0000000000001 ;
      else
        tolerance = 0.0000000000000000001 ;
    }

    // heuristic: horizon for reasonable precision - we assume that no one in their right
    // minds would want a negative horizon ;)

    real_type max_pole = .00000000000000000001 ;
    if ( spline_degree > 1 )
      max_pole = fabs ( vspline_constants::precomputed_poles [ spline_degree ] [ 0 ] ) ;
    if ( smoothing > max_pole )
      max_pole = smoothing ;

    if ( _horizon < 0 )
      horizon = ceil ( log ( tolerance ) / log ( max_pole ) ) ; // TODO what if tolerance == 0.0?
    else
      horizon = _horizon ; // whatever the user specifies

    // first, calculate all the various shapes and sizes used internally
    setup_metrics ( headroom ) ;
//     std::cout << "container shape: " << container_shape << std::endl ;

    // now either adopt external memory or allocate memory for the coefficients
    if ( _space.hasData() )
    {
      // caller has provided space for the coefficient array. This space has to
      // be at least as large as the container_shape we have determined
      // to make sure it's compatible with the other parameters
      if ( ! ( allGreaterEqual ( _space.shape() , container_shape ) ) )
        throw shape_mismatch ( "the intended container shape does not fit into the shape of the storage space passed in" ) ;
      // if the shape matches, we adopt the data in _space;
      // since 'container' was default-constructed, assignment results in a view
      // to the data in _space, not in copying the data. This means that if the data
      // _space refers to change or are deallocated, the bspline will become invalid
      // as well. We take a view to the container_shape-sized subarray only.
      // _coeffs remains uninitialized.
      container = _space.subarray ( shape_type() , container_shape ) ;
    }
    else
    {
      // _space was default-constructed and has no data.
      // in this case we allocate a container array
      array_type help ( container_shape ) ;
      // and swap with the empty default-constructed array _coeffs
      // so that the memory is automatically deallocated when the bspline
      // object is destroyed
      _coeffs.swap ( help ) ;
      container = _coeffs ;
    }
    // finally we set the views to the braced core area and the core area
    coeffs = container.subarray ( left_frame - left_brace ,
                                  left_frame + core_shape + right_brace ) ;
    core = coeffs.subarray ( left_brace , left_brace + core_shape ) ;
  } ;

  /// copy constructor. This will result in a view to the same data and is therefore
  /// lightweight. But the data viewed by this spline must remain accessible.
  
  // TODO this code should handle '_coeffs' via a shared pointer or such,
  // so that the 'last one out' destroys the MultiArray, if the bspline object
  // initially owns the data.

  bspline ( const bspline& other )
  : strategy ( other.strategy ) ,
    container ( other.container ) ,
    coeffs ( other.coeffs ) ,
    core ( other.core ) ,
    spline_degree ( other.spline_degree ) ,
    bcv ( other.bcv ) ,
    tolerance ( other.tolerance ) ,
    smoothing ( other.smoothing ) ,
    braced ( other.braced ) ,
    horizon ( other.horizon ) ,
    left_brace ( other.left_brace ) ,
    right_brace ( other.right_brace ) ,
    left_frame ( other.left_frame ) ,
    right_frame ( other.right_frame ) ,
    container_shape ( other.container_shape ) ,
    core_shape ( other.core_shape ) ,
    braced_shape ( other.braced_shape )
  { } ;
  
  bspline operator= ( const bspline& other )
  {
    return bspline ( *this ) ;
  }
  
  /// get a bspline object for a single channel of the data. This is also lightweight
  /// and requires the viewed data to remain present as long as the channel view is used.
  /// the channel view inherits all metrics from it's parent, only the MultiArrayViews
  /// to the data are different.
  
  channel_view_type get_channel_view ( const int & channel )
  {
    assert ( channel < channels ) ;
    
    real_type * base = (real_type*) ( container.data() ) ;
    base += channel ;
    auto stride = container.stride() ;
    stride *= channels ;
    
    MultiArrayView < dimension , real_type >
      channel_container ( container.shape() , stride , base ) ;

    return channel_view_type ( core_shape , 
                               spline_degree ,
                               bcv ,
                               strategy ,
                               horizon ,
                               tolerance ,
                               smoothing ,
                               0 ,
                               channel_container // coefficient storage to 'adopt'
                             ) ;
  } ;

  /// prefilter converts the knot point data in the 'core' area into b-spline
  /// coefficients. Depending on the strategy chosen in the b-spline object's
  /// constructor, bracing/framing may be applied. Even if the degree of the
  /// spline is zero or one, prefilter() should be called because it also
  /// performs the bracing, if any, which may still be needed if the spline
  /// is 'shifted' - unless the stratgey is UNBRACED, of course.
  ///
  /// If data are passed in, they have to have precisely the shape
  /// we have set up in core (_core_shape passed into the constructor).
  /// These data will then be used in place of any data present in the
  /// bspline object to calculate the coefficients. They won't be looked at
  /// after prefilter() terminates, so it's safe to pass in some MultiArrayView
  /// which is destroyed after the call to prefilter().

  void prefilter ( view_type data = view_type() ) ///< view to knot point data to use instead of 'core'
  {
    if ( data.hasData() )
    {
      // if the user has passed in data, they have to have precisely the shape
      // we have set up in core (_core_shape passed into the constructor).
      // This can have surprising effects if the container array isn't owned by the
      // spline but constitutes a view to data kept elsewhere (by passing _space to the
      // constructor).
      if ( data.shape() != core_shape )
        throw shape_mismatch
         ( "when passing data to prefilter, they have to have precisely the core's shape" ) ;
      if ( strategy == EXPLICIT )
      {
        // the explicit scheme requires the data and frame to be together in the
        // containing array, so we have to copy the data into the core.
        core = data ;
      }
      // the other strategies can move the data from 'data' into the spline's memory
      // during coefficient generation, so we needn't copy them in first.
    }
    else
    {
      // otherwise, we assume data are already in 'core' and we operate in-place
      // note, again, the semantics of the assignment here: since 'data' has no data,
      // the assignment results in 'adopting' the data in core rather than copying them
      data = core ;
    }

    // per default the output will be braced. This does require the output
    // array to be sufficiently larger than the input; class bracer has code
    // to provide the right sizes

    bracer<view_type> br ;

    // for the explicit scheme, we use boundary condition 'guess' which tries to
    // provide a good guess for the initial coefficients with a small computational
    // cost. using zero-padding instead introduces a sharp discontinuity at the
    // margins, which we want to avoid.

    bcv_type explicit_bcv ( GUESS ) ;

    switch ( strategy )
    {
      case UNBRACED:
        // only call the solver, don't do any bracing. If necessary, bracing can be
        // aplied later by a call to brace() - provided the bspline object has space
        // for the brace.
        solve < view_type , view_type , math_type >
              ( data ,
                core ,
                bcv ,
                spline_degree ,
                tolerance ,
                smoothing
              ) ;
        break ;
      case BRACED:
        // solve first, passing in BC codes to pick out the appropriate functions to
        // calculate the initial causal and anticausal coefficient, then brace result.
        // note how, just as in brace(), the whole frame is filled, which may be more
        // than is strictly needed by the evaluator.
        solve < view_type , view_type , math_type >
              ( data ,
                core ,
                bcv ,
                spline_degree ,
                tolerance ,
                smoothing
              ) ;
        // using the more general code here now, since the frame may be larger
        // than strictly necessary for the given spline degree due to a request
        // for additional headroom
        for ( int d = 0 ; d < dimension ; d++ )
          br.apply ( container , bcv[d] , left_frame[d] , right_frame[d] , d ) ;
        break ;
      case EXPLICIT:
        // first fill frame using BC codes passed in, then solve with BC code GUESS
        // this automatically fills the brace as well, since it's part of the frame.
        // TODO: the values in the frame will not come out precisely the same as they
        // would by filling the brace after the coefficients have been calculated.
        // The difference will be larger towards the margin of the frame, and we assume
        // that due to the small support of the evaluation the differences near the
        // margin of the core data will be negligible, having picked a sufficiently
        // large frame size. This is debatable. If it's a problem, a call to brace()
        // after prefilter() will brace again, now with coefficients from the core.
        for ( int d = 0 ; d < dimension ; d++ )
          br.apply ( container , bcv[d] , left_frame[d] , right_frame[d] , d ) ;
        solve < view_type , view_type , math_type >
              ( container ,
                container ,
                explicit_bcv ,
                spline_degree ,
                tolerance ,
                smoothing
              ) ;
        break ;
      case MANUAL:
        // like EXPLICIT, but don't apply a frame, assume a frame was applied
        // by external code. process whole container with GUESS BC. For cases
        // where the frame can't be constructed by applying any of the stock bracing
        // modes. Note that if any data were passed into this routine, in this case
        // they will be silently ignored (makes no sense overwriting the core after
        // having manually framed it in some way)
        solve < view_type , view_type , math_type >
              ( container ,
                container ,
                explicit_bcv ,
                spline_degree ,
                tolerance ,
                smoothing
              ) ;
        break ;
    }
  }

  /// if the spline coefficients are already known, they obviously don't need to be
  /// prefiltered. But in order to be used by vspline's evaluation code, they need to
  /// be 'braced' - the 'core' coefficients have to be surrounded by more coeffcients
  /// covering the support the evaluator needs to operate without bounds checking
  /// inside the spline's defined range. brace() performs this operation. brace()
  /// assumes the bspline object has been set up with the desired initial parameters,
  /// so that the boundary conditions and metrics are already known and storage is
  /// available. If brace() is called with an empty view (or without parameters),
  /// it assumes the coefficients are in the spline's core already and simply
  /// fills in the 'empty' space around them. If data are passed to brace(), they
  /// have to be the same size as the spline's core and are copied into the core
  /// before the bracing is applied.

  void brace ( view_type data = view_type() ) ///< view to knot point data to use instead of 'core'
  {
    if ( data.hasData() )
    {
      // if the user has passed in data, they have to have precisely the shape
      // we have set up in core

      if ( data.shape() != core_shape )
        throw shape_mismatch
         ( "when passing data to prefilter, they have to have precisely the core's shape" ) ;
      
      // we copy the data into the core
      core = data ;
    }

    // we use class bracer to do the work, creating the brace for all axes in turn

    bracer<view_type> br ;

    for ( int d = 0 ; d < dimension ; d++ )
      br.apply ( container , bcv[d] , left_frame[d] , right_frame[d] , d ) ;
  }

  /// overloaded constructor for 1D splines. This is useful because if we don't
  /// provide it, the caller would have to pass TinyVector < T , 1 > instead of T
  /// for the shape and the boundary condition.
  
  bspline ( long _core_shape ,                      ///< shape of knot point data
            int _spline_degree = 3 ,                ///< spline degree with reasonable default
            bc_code _bc = MIRROR ,                  ///< boundary conditions and common default
            prefilter_strategy _strategy = BRACED , ///< default strategy is the 'implicit' scheme
            int _horizon = -1 ,                     ///< width of frame for explicit scheme
            double _tolerance = -1.0 ,              ///< acceptable error (relative to unit pulse)
            double _smoothing = 0.0 ,               ///< apply smoothing to data before prefiltering
            int headroom = 0 ,                      ///< additional headroom, for 'shifting'
            view_type _space = view_type()          ///< coefficient storage to 'adopt'
          )
  :bspline ( TinyVector < long , 1 > ( _core_shape ) ,
             _spline_degree ,
             bcv_type ( _bc ) ,
             _strategy ,
             _horizon ,
             _tolerance ,
             _smoothing ,
             headroom ,
             _space
           )
  {
    static_assert ( _dimension == 1 , "bspline: 1D constructor only usable for 1D splines" ) ;
  } ;

  /// shift will change the interpretation of the data in a bspline object.
  /// d is taken as a difference to add to the current spline degree. The coefficients
  /// remain the same, but creating an evaluator from the shifted spline will make
  /// the evaluator produce data *as if* the coefficients were those of a spline
  /// of the changed order. Shifting with positive d will efectively blur the
  /// interpolated signal, shifting with negative d will sharpen it.
  /// For shifting to work, the spline has to have enough 'headroom', meaning that
  /// spline_degree + d, the new spline degree, has to be greater or equal to 0
  /// and smaller than the largest supported spline degree (lower twenties) -
  /// and, additionally, there has to bee a wide-enough brace to allow evaluation
  /// with the wider kernel of the higher-degree spline's reconstruction filter.
  /// So if a spline is set up with degree 0 and shifted to degree 5, it has to be
  /// constructed with an additional headroom of 3 (see the constructor).
  
  // TODO consider moving the concept of shifting to class evaluator

  int shift ( int d )
  {
    int new_degree = spline_degree + d ;
    if ( new_degree < 0 || new_degree > 24 )
      return 0 ;

    bracer<view_type> br ;
    shape_type new_left_brace = br.left_corner ( bcv , new_degree ) ;
    shape_type new_right_brace = br.right_corner ( bcv , new_degree ) ;
    if (    allLessEqual ( new_left_brace , left_frame )
             && allLessEqual ( new_right_brace , right_frame ) )
    {
      // perform the shift
      spline_degree = new_degree ;
      left_brace = new_left_brace ;
      right_brace = new_right_brace ;
      braced_shape = core_shape + left_brace + right_brace ;

      shape_type coefs_offset = left_frame - new_left_brace ;
      coeffs.reset() ;
      coeffs = container.subarray ( coefs_offset , coefs_offset + braced_shape ) ;
    }
    else
    {
      // can't shift
      std::cout << "can't shift" << std::endl ;
      d = 0 ;
    }

    return d ;
  }

  /// helper function to << a bspline object to an ostream

  friend ostream& operator<< ( ostream& osr , const bspline& bsp )
  {
    osr << "dimension:................... " << bsp.dimension << endl ;
    osr << "degree:...................... " << bsp.spline_degree << endl ;
    osr << "boundary conditions:......... " ;
    for ( auto bc : bsp.bcv )
      osr << " " << bc_name [ bc ] ;
    osr << endl ;
    osr << endl ;
    osr << "shape of container array:.... " << bsp.container.shape() << endl ;
    osr << "shape of braced coefficients: " << bsp.coeffs.shape() << endl ;
    osr << "shape of core:............... " << bsp.core.shape() << endl ;
    osr << "braced:...................... " << ( bsp.braced ? std::string ( "yes" ) : std::string ( "no" ) ) << endl ;
    osr << "left brace:.................. " << bsp.left_brace << endl ;
    osr << "right brace:................. " << bsp.right_brace << endl ;
    osr << "left frame:.................. " << bsp.left_frame << endl ;
    osr << "right frame:................. " << bsp.right_frame << endl ;
    osr << ( bsp._coeffs.hasData() ? "bspline object owns data" : "data are owned externally" ) << endl ;
    osr << "container base adress:....... " << bsp.container.data() << endl ;
    return osr ;
  }

} ;

} ; // end of namespace vspline

#endif // VSPLINE_BSPLINE_H
