/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file common.h

    \brief definitions common to all files in this project, utility code
    
    This file contains
    
    - a traits class fixing the simdized types used for vectorized code
    
    - exceptions used throughout vspline
*/

#ifndef VSPLINE_COMMON
#define VSPLINE_COMMON

#include <vigra/multi_array.hxx>

#ifdef USE_VC

#include <Vc/Vc>

#endif

namespace vspline
{

/// This enumeration is used for codes connected to boundary conditions. There are
/// two aspects to boundary conditions: During prefiltering, if the implicit scheme is used,
/// the initial causal and anticausal coefficients have to be calculated in a way specific to
/// the chosen boundary conditions. Bracing, both before prefiltering when using the explicit
/// scheme, and after prefiltering when using the implicit scheme, also needs these codes to
/// pick the appropriate extrapolation code to extend the knot point data/coefficients beyond
/// the core array.

typedef enum { 
  MIRROR ,    ///< mirror on the bounds, so that f(-x) == f(x)
  PERIODIC,   ///< periodic boundary conditions
  REFLECT ,   ///< reflect, so  that f(-1) == f(0) (mirror between bounds)
  NATURAL,    ///< natural boundary conditions, f(-x) + f(x) == 2 * f(0)
  CONSTANT ,  ///< clamp. used for framing, with explicit prefilter scheme
  ZEROPAD ,   ///< used for boundary condition, bracing
  IDENTITY ,  ///< used as solver argument, mostly internal use
  GUESS ,     ///< used with EXPLICIT scheme to keep margin errors low
  SPHERICAL , ///< use for spherical panoramas, y axis
} bc_code;

/// bc_name is for diagnostic output of bc codes

const std::string bc_name[] =
{
  "MIRROR" ,
  "PERIODIC",
  "REFLECT" ,
  "NATURAL",
  "CONSTANT" ,
  "ZEROPAD" ,
  "IDENTITY" ,
  "GUESS" ,
  "SPHERICAL" ,
} ;

#ifdef USE_VC

/// using definition for the 'elementary type' of a type via vigra's
/// ExpandElementResult mechanism.

template < class T >
using ET = typename vigra::ExpandElementResult < T > :: type ;

/// struct vector_traits is used throughout vspline to determine simdized
/// data types. While inside the vspline code base, vsize, the number of
/// elementary type members in a simdized type, is kept variable in form of
/// the template argument _vsize, in code using vspline you'd expect to find
/// a program-wide definition deriving the number of elements from the
/// type commonly used for arithmetics, like
///
/// #define VSIZE (vspline::vector_traits<float>::size)
///
/// If other elementary types are used, their simdized types are fixed to use
/// the same number of elements by using the program-wide number of elements:
///
/// typedef vspline::vector_traits<double,VSIZE>::ele_v double_v_type ;
///
/// typedef vigra::TinyVector < int , 3 > triplet_type ;
///
/// typedef vspline::vector_traits<triplet_type,VSIZE>::type triplet_v_type ;
///
/// using a default _vsize of twice the elementary type's Vc::Vector's size
/// works best on my system, but may not be optimal elsewhere.

template < class T ,
           int _vsize = 2 * Vc::Vector < ET<T> > :: Size >
struct vector_traits
{
  enum { size = _vsize } ;

  enum { dimension = vigra::ExpandElementResult < T > :: size } ;

  typedef ET<T> ele_type ;

  typedef vigra::TinyVector < ele_type , dimension > nd_ele_type ;
  
  typedef Vc::SimdArray < ele_type , size > ele_v ;
  
  typedef vigra::TinyVector < ele_v , dimension > nd_ele_v ;
  
  typedef vigra::TinyVector < ele_v , dimension > type ;
  
//   typedef typename std::conditional
//           < dimension == 1 ,
//             ele_v ,          
//             vigra::TinyVector < ele_v , dimension >
//           > :: type type ;
} ;

#endif

// TODO The exceptions need some work. My use of exceptions is a bit sketchy...

/// for interfaces which need specific implementations we use:

struct not_implemented
: std::invalid_argument
{
  not_implemented ( const char * msg )
  : std::invalid_argument ( msg ) { }  ;
} ;

/// dimension-mismatch is thrown if two arrays have different dimensions
/// which should have the same dimensions.

struct dimension_mismatch
: std::invalid_argument
{
  dimension_mismatch ( const char * msg )
  : std::invalid_argument ( msg ) { }  ;
} ;

/// shape mismatch is the exception which is thrown if the shapes of
/// an input array and an output array do not match.

struct shape_mismatch
: std::invalid_argument
{
  shape_mismatch  ( const char * msg )
  : std::invalid_argument ( msg ) { }  ;
} ;

/// exception which is thrown if an opertion is requested which vspline
/// does not support

struct not_supported
: std::invalid_argument
{
  not_supported  ( const char * msg )
  : std::invalid_argument ( msg ) { }  ;
} ;

/// out_of_bounds is thrown by mapping mode REJECT for out-of-bounds coordinates
/// this exception is left without a message, it only has a very specific application,
/// and there it may be thrown often, so we don't want anything slowing it down.

struct out_of_bounds
{
} ;

/// exception which is thrown when an assiging an rvalue which is larger than
/// what the lvalue can hold

struct numeric_overflow
: std::invalid_argument
{
  numeric_overflow  ( const char * msg )
  : std::invalid_argument ( msg ) { }  ;
} ;

} ; // end of namespace vspline

#endif // VSPLINE_COMMON
