/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015, 2016 by Kay F. Jahnke                     */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/// \file complex.cc
///
/// \brief demonstrate use of b-spline over std::complex data
///
/// vspline handles std::complex data like pairs of the complex
/// type's value_type, and uses a vigra::TinyVector of two
/// simdized value_types as the vectorized type.
///
/// compile: clang++ -std=c++11 -march=native -o complex -O3 -pthread -DUSE_VC complex.cc -lvigraimpex -lVc

#include <iomanip>
#include <assert.h>
#include <complex>
#include <vspline/multithread.h>
#include <vspline/vspline.h>

int main ( int argc , char * argv[] )
{
  vspline::bspline < std::complex < float > , 1 > bsp ( 100000 , 3 , vspline::MIRROR ) ;
  auto v1 = bsp.core ;
  v1 [ 50000 ] = std::complex<float> ( 1.0 + 1i ) ;
  bsp.prefilter() ;

  typedef vspline::evaluator < float , std::complex<float> > ev_type ;
  
  ev_type ev ( bsp ) ;
  for ( float k = 49999.0 ; k < 50001.0 ; k += .1 )
  {
    std::cout << "ev(" << k << ") = " << ev(k) << std::endl ;
  }
  
#ifdef USE_VC

  for ( float k = 49999.0 ; k < 50001.0 ; k += .1 )
  {
    // feed the evaluator with vectors, just to show off. Note how the
    // result appears as a vigra::TinyVector of two ev_type::out_v.
    typename ev_type::in_v vk ( k ) ;
    std::cout << "ev(" << vk << ") = " << ev(vk) << std::endl ;
  }
  
#endif
}
