/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015, 2016 by Kay F. Jahnke                     */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/// \file splinus.cc
///
/// \brief compare a periodic b-spline with a sine
/// 
/// This is a deliberately trivial example using a periodic b-spline
/// over just two values: 1 and -1. This spline is used to approximate
/// a sine function. You pass the spline's desired degree on the command
/// line. Next you enter a number (interpreted as degrees) and the program
/// will output the sine and the 'splinus' of the given angle.
/// As you can see when playing with higher degrees, the higher the spline's
/// degree, the closer the match with the sine. So apart from serving as a
/// very simple demonstration of using a 1D periodic b-spline, it teaches us
/// that a periodic b-spline can approximate a sine.
/// To show off, we use long double as the spline's data type - but we can
/// only do so without using Vc, since long doubles can't be vectorized.
///
/// compile with: clang++ -pthread -O3 -std=c++11 splinus.cc -o splinus

// if USE_VC was defined for compiling this example, we undefine it, since
// long double can't be vectorized.

#undef USE_VC

#include <assert.h>
#include <vspline/vspline.h>
#include <vspline/map.h>

int main ( int argc , char * argv[] )
{
  assert ( argc > 1 ) ;
  
  int degree = std::atoi ( argv[1] ) ;
  
  assert ( degree >= 0 && degree < 25 ) ;
  
  // create the bspline object
  
  vspline::bspline < long double ,  // spline's data type
                     1 >            // one dimension
    bsp ( 2 ,                   // two values
          degree ,              // degree as per command line
          vspline::PERIODIC ,   // periodic boundary conditions
          vspline::BRACED ,     // implicit scheme, bracing coeffs
          -1 ,                  // horizon (unused w. BRACED)
          0.0 ) ;               // no tolerance
          
  // the bspline object's 'core' is a MultiArrayView to the knot point
  // data, which we set one by one for this simple example:
  
  bsp.core[0] = 1.0L ;
  bsp.core[1] = -1.0L ;
  
  // now we prefilter the data
  
  bsp.prefilter() ;
  
  // the spline is now ready for use, we create an evaluator
  // to obtain interpolated values
  
  typedef vspline::evaluator < long double , long double > ev_type ;
  
  ev_type ev ( bsp ) ; // from the bspline object we just made

  // we want to map the incoming coordinates into the defined range.
  // Here we go all the way, using a vspline::mapper object, which
  // is overkill for the simple 1D operation, but it's instructive.

  // First define the 'gate' type - this type 'treats' a single value

  typedef vspline::periodic_gate_type < long double > pg_type ;

  // next we define the mapper type, which is n-dimensional, but since
  // we only use one dimension, it's a mere shell around the gate type

  typedef vspline::mapper < long double ,  // coordinate type
                            1 ,            // vector size
                            pg_type        // the single gate type used here
                          > mapper_type ;
  
  // now we create the mapper, passing the periodic gate with the desired start
  // and end point to the mapper's constructor

  mapper_type m ( pg_type ( 0.0L , 2.0L ) ) ;
  
  // finally we chain the mapper and the evaluator to create a functor
  // which first passes the coordinate through the 'gate', mapping it to
  // [ 0 : 2 ], and then evaluates at the mapped coordinate:

  auto periodic_ev = vspline::chain < mapper_type , ev_type > ( m , ev ) ;
  
  while ( true )
  {
    std::cout << " > " ;
    long double x ;
    std::cin >> x ;                       // get an angle
    long double xs = x * M_PI / 180.0L ;  // sin() uses radians 
    long double xx = x / 180.0L - .5L ;   // 'splinus' has period 2 and is shifted .5
    
    // finally we can produce both results. Note how we can use periodic_ev,
    // the combination of gate and evaluator, like an ordinary function.

    std::cout << "sin(" << xs << ") = "
              << sin(xs)
              << " splinus(" << xx << ") = "
              << periodic_ev ( xx )
              << " difference: "
              << sin(xs) - periodic_ev(xx)
              << std::endl ;

    // for this simple example we might as well have used the gate directly:

    assert ( m ( xx ) == pg_type ( 0.0L , 2.0L ) ( xx ) ) ;
  }
}
