/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file poles.h

    \brief precalculated prefilter poles and basis function values

    The contents of this file below the comments can be generated
    using prefilter_poles.cc

    While the precalculated basis function values can be generated in long double
    precision (with code in basis.h), the filter poles are calculated using
    gsl and BLAS, which provide only double precision.
    
    The values defined here are used in several places in vspline. They are
    precomputed because calculating them when needed can be (potentially very)
    expensive, and providing them by definitions evaluated at compile time
    slows compilation.
    
    The set of values provided here is sufficient to calculate the b-spline
    basis function for all spline degrees for arbitrary arguments (including
    the spline's derivatives) - see basis.h. The poles are needed for prefiltering.
*/

#ifndef VSPLINE_POLES_H

namespace vspline_constants {

const long double K0[] = {
 1L ,   // basis(0)
 } ; 
const long double K1[] = {
 1L ,   // basis(0)
 0.5L ,   // basis(0.5)
 } ; 
const long double K2[] = {
 0.75L ,   // basis(0)
 0.5L ,   // basis(0.5)
 0.125L ,   // basis(1)
 } ; 
const double Poles_2[] = {
-0.17157287525381015314 ,
} ;
const long double K3[] = {
 0.66666666666666666668L ,   // basis(0)
 0.47916666666666666666L ,   // basis(0.5)
 0.16666666666666666667L ,   // basis(1)
 0.020833333333333333334L ,   // basis(1.5)
 } ; 
const double Poles_3[] = {
-0.26794919243112280682 ,
} ;
const long double K4[] = {
 0.59895833333333333332L ,   // basis(0)
 0.45833333333333333334L ,   // basis(0.5)
 0.19791666666666666667L ,   // basis(1)
 0.041666666666666666668L ,   // basis(1.5)
 0.0026041666666666666667L ,   // basis(2)
 } ; 
const double Poles_4[] = {
-0.36134122590021944266 ,
-0.013725429297339164503 ,
} ;
const long double K5[] = {
 0.55000000000000000001L ,   // basis(0)
 0.4380208333333333333L ,   // basis(0.5)
 0.21666666666666666667L ,   // basis(1)
 0.061718750000000000001L ,   // basis(1.5)
 0.0083333333333333333337L ,   // basis(2)
 0.00026041666666666666668L ,   // basis(2.5)
 } ; 
const double Poles_5[] = {
-0.43057534709997430378 ,
-0.043096288203264443428 ,
} ;
const long double K6[] = {
 0.51102430555555555553L ,   // basis(0)
 0.41944444444444444449L ,   // basis(0.5)
 0.22879774305555555554L ,   // basis(1)
 0.079166666666666666666L ,   // basis(1.5)
 0.015668402777777777778L ,   // basis(2)
 0.001388888888888888889L ,   // basis(2.5)
 2.170138888888888889e-05L ,   // basis(3)
 } ; 
const double Poles_6[] = {
-0.48829458930304570075 ,
-0.081679271076237264237 ,
-0.0014141518083258114435 ,
} ;
const long double K7[] = {
 0.47936507936507936512L ,   // basis(0)
 0.4025964161706349206L ,   // basis(0.5)
 0.23630952380952380952L ,   // basis(1)
 0.094024367559523809525L ,   // basis(1.5)
 0.023809523809523809525L ,   // basis(2)
 0.0033776661706349206347L ,   // basis(2.5)
 0.00019841269841269841271L ,   // basis(3)
 1.5500992063492063493e-06L ,   // basis(3.5)
 } ; 
const double Poles_7[] = {
-0.5352804307964414976 ,
-0.12255461519232610512 ,
-0.0091486948096082820747 ,
} ;
const long double K8[] = {
 0.45292096819196428568L ,   // basis(0)
 0.38737599206349206352L ,   // basis(0.5)
 0.24077768477182539682L ,   // basis(1)
 0.10647321428571428571L ,   // basis(1.5)
 0.032126968625992063494L ,   // basis(2)
 0.0061259920634920634923L ,   // basis(2.5)
 0.00063476562499999999998L ,   // basis(3)
 2.4801587301587301589e-05L ,   // basis(3.5)
 9.6881200396825396832e-08L ,   // basis(4)
 } ; 
const double Poles_8[] = {
-0.57468690924881216109 ,
-0.16303526929727354955 ,
-0.023632294694844447475 ,
-0.00015382131064169135559 ,
} ;
const long double K9[] = {
 0.43041776895943562614L ,   // basis(0)
 0.3736024025676532187L ,   // basis(0.5)
 0.24314925044091710759L ,   // basis(1)
 0.1168385769744819224L ,   // basis(1.5)
 0.040255731922398589063L ,   // basis(2)
 0.0094531293058311287482L ,   // basis(2.5)
 0.0013833774250440917108L ,   // basis(3)
 0.00010588576974481922398L ,   // basis(3.5)
 2.7557319223985890654e-06L ,   // basis(4)
 5.3822889109347442683e-09L ,   // basis(4.5)
 } ; 
const double Poles_9[] = {
-0.60799738916862233751 ,
-0.20175052019315406482 ,
-0.04322260854048156492 ,
-0.0021213069031808251541 ,
} ;
const long double K10[] = {
 0.41096264282441854056L ,   // basis(0)
 0.36109843474426807762L ,   // basis(0.5)
 0.24406615618885719797L ,   // basis(1)
 0.12543871252204585538L ,   // basis(1.5)
 0.047983348920442019401L ,   // basis(2)
 0.013183421516754850087L ,   // basis(2.5)
 0.0024532852307408785274L ,   // basis(3)
 0.00027915564373897707232L ,   // basis(3.5)
 1.5887978636188271604e-05L ,   // basis(4)
 2.7557319223985890654e-07L ,   // basis(4.5)
 2.6911444554673721342e-10L ,   // basis(5)
 } ; 
const double Poles_10[] = {
-0.63655066396958059904 ,
-0.23818279837754796624 ,
-0.065727033228304657109 ,
-0.0075281946755491966489 ,
-1.6982762823274620556e-05 ,
} ;
const long double K11[] = {
 0.39392556517556517558L ,   // basis(0)
 0.34970223188744306906L ,   // basis(0.5)
 0.24396028739778739779L ,   // basis(1)
 0.13256116543210659421L ,   // basis(1.5)
 0.055202020202020202023L ,   // basis(2)
 0.017163149607531321399L ,   // basis(2.5)
 0.0038238786676286676285L ,   // basis(3)
 0.00057128626126327135446L ,   // basis(3.5)
 5.1006092672759339424e-05L ,   // basis(4)
 2.1667994232691498315e-06L ,   // basis(4.5)
 2.5052108385441718776e-08L ,   // basis(5)
 1.2232474797578964246e-11L ,   // basis(5.5)
 } ; 
const double Poles_11[] = {
-0.66126606890063921451 ,
-0.27218034929481393913 ,
-0.089759599793708844118 ,
-0.016669627366234951449 ,
-0.00051055753444649576434 ,
} ;
const long double K12[] = {
 0.37884408454472999147L ,   // basis(0)
 0.33927295023649190319L ,   // basis(0.5)
 0.24313091801014469471L ,   // basis(1)
 0.13845146655042488376L ,   // basis(1.5)
 0.061867668009041325489L ,   // basis(2)
 0.021268582401394901395L ,   // basis(2.5)
 0.0054581869256967252307L ,   // basis(3)
 0.00099847474413446635658L ,   // basis(3.5)
 0.00012091392059187537162L ,   // basis(4)
 8.5239798781465448132e-06L ,   // basis(4.5)
 2.7086165069699140876e-07L ,   // basis(5)
 2.0876756987868098981e-09L ,   // basis(5.5)
 5.0968644989912351028e-13L ,   // basis(6)
 } ; 
const double Poles_12[] = {
-0.68286488419809487915 ,
-0.30378079328817336746 ,
-0.11435052002714780894 ,
-0.028836190198661435652 ,
-0.0025161662172618224839 ,
-1.883305645063344802e-06 ,
} ;
const long double K13[] = {
 0.36537086948545281884L ,   // basis(0)
 0.32968987958591001189L ,   // basis(0.5)
 0.24178841798633465302L ,   // basis(1)
 0.14331501747174208366L ,   // basis(1.5)
 0.067974967258821425491L ,   // basis(2)
 0.0254044062949849888L ,   // basis(2.5)
 0.0073122366959172514726L ,   // basis(3)
 0.0015671731081656330546L ,   // basis(3.5)
 0.00023762984700484700481L ,   // basis(4)
 2.3492285420207986942e-05L ,   // basis(4.5)
 1.3133086049752716419e-06L ,   // basis(5)
 3.125375747123929632e-08L ,   // basis(5.5)
 1.6059043836821614601e-10L ,   // basis(6)
 1.960332499612013501e-14L ,   // basis(6.5)
 } ; 
const double Poles_13[] = {
-0.701894251817016257 ,
-0.33310723293052579841 ,
-0.13890111319434489401 ,
-0.043213866740361948915 ,
-0.0067380314152448743045 ,
-0.00012510011321441739246 ,
} ;
const long double K14[] = {
 0.35323915669918929845L ,   // basis(0)
 0.32085024502063192543L ,   // basis(0.5)
 0.24008299041558734203L ,   // basis(1)
 0.14732180094624291054L ,   // basis(1.5)
 0.073541032564067060978L ,   // basis(2)
 0.029499800232377117299L ,   // basis(2.5)
 0.009341081854512256905L ,   // basis(3)
 0.002275919650051594496L ,   // basis(3.5)
 0.00041109051149372196722L ,   // basis(4)
 5.2046374591017448155e-05L ,   // basis(4.5)
 4.2229561084936041826e-06L ,   // basis(5)
 1.8776463468923786384e-07L ,   // basis(5.5)
 3.3486357751247422931e-09L ,   // basis(6)
 1.1470745597729724715e-11L ,   // basis(6.5)
 7.0011874986143339324e-16L ,   // basis(7)
 } ; 
const double Poles_14[] = {
-0.71878378723766189751 ,
-0.36031907191881451524 ,
-0.16303351479903732679 ,
-0.059089482194828991946 ,
-0.013246756734847169382 ,
-0.00086402404095337124838 ,
-2.0913096775274000322e-07 ,
} ;
const long double K15[] = {
 0.34224026135534072046L ,   // basis(0)
 0.31266660625176080971L ,   // basis(0.5)
 0.23812319491070731152L ,   // basis(1)
 0.15061194980399698684L ,   // basis(1.5)
 0.078595253866748575742L ,   // basis(2)
 0.033503802571649835525L ,   // basis(2.5)
 0.011502274487496875064L ,   // basis(3)
 0.003117493948498863913L ,   // basis(3.5)
 0.00064854900635323915745L ,   // basis(4)
 9.9440249438946462506e-05L ,   // basis(4.5)
 1.057200426826749578e-05L ,   // basis(5)
 7.0683979027987963181e-07L ,   // basis(5.5)
 2.5045990654456262921e-08L ,   // basis(6)
 3.348642542939324287e-10L ,   // basis(6.5)
 7.6471637318198164765e-13L ,   // basis(7)
 2.3337291662047779775e-17L ,   // basis(7.5)
 } ; 
const double Poles_15[] = {
-0.73387257168597164192 ,
-0.3855857342780184549 ,
-0.18652010845105168602 ,
-0.075907592047656735623 ,
-0.021752065796541687759 ,
-0.0028011514820764091618 ,
-3.0935680451474410063e-05 ,
} ;
const long double K16[] = {
 0.33220826914249586032L ,   // basis(0)
 0.30506442781494322298L ,   // basis(0.5)
 0.23598831687663609049L ,   // basis(1)
 0.15330093144015230863L ,   // basis(1.5)
 0.083172975045518980468L ,   // basis(2)
 0.03738103391018481751L ,   // basis(2.5)
 0.013757630909488189399L ,   // basis(3)
 0.0040808725321077028254L ,   // basis(3.5)
 0.00095448286788948239937L ,   // basis(4)
 0.00017072700505627712969L ,   // basis(4.5)
 2.2348950637818187112e-05L ,   // basis(5)
 2.0041660421228046889e-06L ,   // basis(5.5)
 1.1074718796168506873e-07L ,   // basis(6)
 3.131465753406890973e-09L ,   // basis(6.5)
 3.1393546448057462799e-11L ,   // basis(7)
 4.7794773323873852978e-14L ,   // basis(7.5)
 7.2929036443899311795e-19L ,   // basis(8)
 } ; 
const double Poles_16[] = {
-0.74743238775188380885 ,
-0.40907360475830745195 ,
-0.2092287193405746315 ,
-0.093254718980160661301 ,
-0.031867706120390963676 ,
-0.0062584067851372366872 ,
-0.00030156536330664312833 ,
-2.3232486364235544612e-08 ,
} ;
const long double K17[] = {
 0.32300939415699870668L ,   // basis(0)
 0.29797995870819162778L ,   // basis(0.5)
 0.23373674923065111L ,   // basis(1)
 0.15548403615015999844L ,   // basis(1.5)
 0.087311640770182303119L ,   // basis(2)
 0.041108064309116914771L ,   // basis(2.5)
 0.016073921990964784645L ,   // basis(3)
 0.0051528238735766806875L ,   // basis(3.5)
 0.0013308125721335362832L ,   // basis(4)
 0.00027040492583018919549L ,   // basis(4.5)
 4.1821549694989869669e-05L ,   // basis(5)
 4.695715379871064023e-06L ,   // basis(5.5)
 3.5643941839232455477e-07L ,   // basis(6)
 1.6314974698479856617e-08L ,   // basis(6.5)
 3.6845271901099787809e-10L ,   // basis(7)
 2.7700195120810122023e-12L ,   // basis(7.5)
 2.8114572543455207634e-15L ,   // basis(8)
 2.144971660114685641e-20L ,   // basis(8.5)
 } ; 
const double Poles_17[] = {
-0.75968322407197097501 ,
-0.43093965318021570932 ,
-0.23108984359938430919 ,
-0.11082899331622909911 ,
-0.043213911456682692347 ,
-0.011258183689472329655 ,
-0.0011859331251521279364 ,
-7.6875625812547303262e-06 ,
} ;
const long double K18[] = {
 0.31453440085864671822L ,   // basis(0)
 0.29135844665108330336L ,   // basis(0.5)
 0.2314117793664616011L ,   // basis(1)
 0.15724011346206745634L ,   // basis(1.5)
 0.091048500593391361557L ,   // basis(2)
 0.044670474960158529868L ,   // basis(2.5)
 0.018422928690498247476L ,   // basis(3)
 0.0063191164101958155308L ,   // basis(3.5)
 0.0017772776557432943289L ,   // basis(4)
 0.00040219803091097442175L ,   // basis(4.5)
 7.1383891069110100448e-05L ,   // basis(5)
 9.5907105586580192779e-06L ,   // basis(5.5)
 9.271047742986201033e-07L ,   // basis(6)
 5.9734083260063868359e-08L ,   // basis(6.5)
 2.2685078926749432357e-09L ,   // basis(7)
 4.0941846266406646114e-11L ,   // basis(7.5)
 2.308349801939754902e-13L ,   // basis(8)
 1.5619206968586226463e-16L ,   // basis(8.5)
 5.958254611429682336e-22L ,   // basis(9)
 } ; 
const double Poles_18[] = {
-0.77080505126463716437 ,
-0.45132873338515144823 ,
-0.25207457469899424707 ,
-0.12841283679297030296 ,
-0.055462967138511676257 ,
-0.017662377684794876992 ,
-0.0030119307290000858941 ,
-0.00010633735588702059982 ,
-2.5812403962584360567e-09 ,
} ;
const long double K19[] = {
 0.3066931017379824246L ,   // basis(0)
 0.28515265744763108603L ,   // basis(0.5)
 0.22904564568118377632L ,   // basis(1)
 0.1586346253388907509L ,   // basis(1.5)
 0.094419295116760105743L ,   // basis(2)
 0.048060545425350700269L ,   // basis(2.5)
 0.020781149371245016366L ,   // basis(3)
 0.0075653834126722674764L ,   // basis(3.5)
 0.0022918668891541334257L ,   // basis(4)
 0.00056895229089948501399L ,   // basis(4.5)
 0.00011341320068077591568L ,   // basis(5)
 1.7663033358559161242e-05L ,   // basis(5.5)
 2.0693993456206894214e-06L ,   // basis(6)
 1.7275247843548983816e-07L ,   // basis(6.5)
 9.4683295350905535817e-09L ,   // basis(7)
 2.987004917810922453e-10L ,   // basis(7.5)
 4.3098159994772440922e-12L ,   // basis(8)
 1.8223814805986014024e-14L ,   // basis(8.5)
 8.2206352466243297175e-18L ,   // basis(9)
 1.5679617398499164042e-23L ,   // basis(9.5)
 } ; 
const double Poles_19[] = {
-0.78094644484628727987 ,
-0.47037281947078746214 ,
-0.27218037628176311449 ,
-0.14585089375766777109 ,
-0.068345906124943789361 ,
-0.025265073344845085518 ,
-0.0059366595910830613492 ,
-0.00050841019468083302468 ,
-1.9154786562122251559e-06 ,
} ;
const long double K20[] = {
 0.29941029032001264032L ,   // basis(0)
 0.27932165599364228926L ,   // basis(0.5)
 0.22666242185748694763L ,   // basis(1)
 0.15972211762658876278L ,   // basis(1.5)
 0.0974575566598727568L ,   // basis(2)
 0.051275465138013302938L ,   // basis(2.5)
 0.023129338338060293147L ,   // basis(3)
 0.0088777091023436491261L ,   // basis(3.5)
 0.0028712400200206135652L ,   // basis(4)
 0.00077261996725682196449L ,   // basis(4.5)
 0.00017015073085024172881L ,   // basis(5)
 3.0008819646690530456e-05L ,   // basis(5.5)
 4.1167033003850903957e-06L ,   // basis(6)
 4.2192794922896485481e-07L ,   // basis(6.5)
 3.0493046656519177393e-08L ,   // basis(7)
 1.4241282646631125569e-09L ,   // basis(7.5)
 3.7354418501332067724e-11L ,   // basis(8)
 4.3098940955120870235e-13L ,   // basis(8.5)
 1.3667861257365780153e-15L ,   // basis(9)
 4.1103176233121648586e-19L ,   // basis(9.5)
 3.9199043496247910105e-25L ,   // basis(10)
 } ; 
const double Poles_20[] = {
-0.79023111767977516351 ,
-0.48819126033675236398 ,
-0.29142160165551617146 ,
-0.16303353479638585388 ,
-0.081648115630934034459 ,
-0.033849479552361630419 ,
-0.0099730290200507193399 ,
-0.0014683217571042010263 ,
-3.7746573197331790075e-05 ,
-2.8679944881725126467e-10 ,
} ;
const long double K21[] = {
 0.29262268723143477922L ,   // basis(0)
 0.2738298047486301248L ,   // basis(0.5)
 0.22428009387883276411L ,   // basis(1)
 0.16054821266164454585L ,   // basis(1.5)
 0.10019429073492722872L ,   // basis(2)
 0.054315966627272970968L ,   // basis(2.5)
 0.025451983263662738633L ,   // basis(3)
 0.010243000848845290252L ,   // basis(3.5)
 0.0035111077726313273026L ,   // basis(4)
 0.0010143045932529873796L ,   // basis(4.5)
 0.000243612424661332394L ,   // basis(5)
 4.7797839244413500002e-05L ,   // basis(5.5)
 7.4865177795402407054e-06L ,   // basis(6)
 9.0756157943914249454e-07L ,   // basis(6.5)
 8.1587909794275973583e-08L ,   // basis(7)
 5.1150819066710363876e-09L ,   // basis(7.5)
 2.0383683775099098269e-10L ,   // basis(8)
 4.4482237420372396468e-12L ,   // basis(8.5)
 4.104700189226971567e-14L ,   // basis(9)
 9.7627580792412901893e-17L ,   // basis(9.5)
 1.9572941063391261232e-20L ,   // basis(10)
 9.3331055943447405012e-27L ,   // basis(10.5)
 } ; 
const double Poles_21[] = {
-0.79876288565466957436 ,
-0.50489153745536197171 ,
-0.30982319641503575092 ,
-0.17988466679726275443 ,
-0.095200812461283090826 ,
-0.043213918440668783183 ,
-0.01504549998728420962 ,
-0.0031720039638856827036 ,
-0.00021990295763158517806 ,
-4.7797646894259869337e-07 ,
} ;
const long double K22[] = {
 0.28627661405538603955L ,   // basis(0)
 0.26864594027689889732L ,   // basis(0.5)
 0.22191207309687150606L ,   // basis(1)
 0.1611512144701082552L ,   // basis(1.5)
 0.10265788953426401334L ,   // basis(2)
 0.057185290104801063607L ,   // basis(2.5)
 0.027736783120003498267L ,   // basis(3)
 0.011649203759035082664L ,   // basis(3.5)
 0.0042065557982618627852L ,   // basis(4)
 0.0012943433274091186101L ,   // basis(4.5)
 0.00033552928198532912351L ,   // basis(5)
 7.2224788646371748003e-05L ,   // basis(5.5)
 1.2671383794748147439e-05L ,   // basis(6)
 1.7682350579090077749e-06L ,   // basis(6.5)
 1.8993891467043433632e-07L ,   // basis(7)
 1.5010206322471487409e-08L ,   // basis(7.5)
 8.1770577437810697864e-10L ,   // basis(8)
 2.7833247876855379198e-11L ,   // basis(8.5)
 5.0557094184087925374e-13L ,   // basis(9)
 3.7315643098318982977e-15L ,   // basis(9.5)
 6.6564259722400510509e-18L ,   // basis(10)
 8.8967913924505732872e-22L ,   // basis(10.5)
 2.1211603623510773867e-28L ,   // basis(11)
 } ; 
const double Poles_22[] = {
-0.80662949916286152963 ,
-0.52057023687190062677 ,
-0.3274164733138280603 ,
-0.19635282650762261869 ,
-0.10887245188483440916 ,
-0.053181604599218119944 ,
-0.021035660929842874001 ,
-0.0057066136460001649564 ,
-0.00072254796507928529137 ,
-1.3458154983225084633e-05 ,
-3.186643260432269507e-11 ,
} ;
const long double K23[] = {
 0.28032619854980754502L ,   // basis(0)
 0.26374269458034057742L ,   // basis(0.5)
 0.21956831005031718209L ,   // basis(1)
 0.16156340331433543452L ,   // basis(1.5)
 0.1048741828768824975L ,   // basis(2)
 0.059888404600676471811L ,   // basis(2.5)
 0.029974159449075470104L ,   // basis(3)
 0.013085403104047330802L ,   // basis(3.5)
 0.0049523097091663721335L ,   // basis(4)
 0.0016124087669444304968L ,   // basis(4.5)
 0.00044731411734139782554L ,   // basis(5)
 0.0001044647630135970384L ,   // basis(5.5)
 2.0225085344373592521e-05L ,   // basis(6)
 3.1828904692399063536e-06L ,   // basis(6.5)
 3.9679866129008683199e-07L ,   // basis(7)
 3.7855233852927286935e-08L ,   // basis(7.5)
 2.6346734890183937921e-09L ,   // basis(8)
 1.2488410498396141086e-10L ,   // basis(8.5)
 3.6338307165683742375e-12L ,   // basis(9)
 5.4959585554808751978e-14L ,   // basis(9.5)
 3.2448470402629826029e-16L ,   // basis(10)
 4.3411473752750814749e-19L ,   // basis(10.5)
 3.868170170630684038e-23L ,   // basis(11)
 4.6112181790240812755e-30L ,   // basis(11.5)
 } ; 
const double Poles_23[] = {
-0.81390562354320794558 ,
-0.53531408371104993726 ,
-0.34423627688965990901 ,
-0.21240466055269885404 ,
-0.12256116098899572098 ,
-0.063602480154273194346 ,
-0.027811662038017159748 ,
-0.0090795953352833073946 ,
-0.0017112714467820973156 ,
-9.5733943500721317005e-05 ,
-1.1936918816067781773e-07 ,
} ;
const long double K24[] = {
 0.27473197352118810147L ,   // basis(0)
 0.25909593388549224613L ,   // basis(0.5)
 0.21725612218406020861L ,   // basis(1)
 0.16181208211791016533L ,   // basis(1.5)
 0.10686656672959712099L ,   // basis(2)
 0.062431425854373209442L ,   // basis(2.5)
 0.032156816325798337903L ,   // basis(3)
 0.014541849599514216045L ,   // basis(3.5)
 0.0057429446266243922923L ,   // basis(4)
 0.0019676174028389475043L ,   // basis(4.5)
 0.00058004996270088237076L ,   // basis(5)
 0.00014563543156618789351L ,   // basis(5.5)
 3.0746018052888292381e-05L ,   // basis(6)
 5.3704036096147168721e-06L ,   // basis(6.5)
 7.6016977670631529332e-07L ,   // basis(7)
 8.4861949009616751487e-08L ,   // basis(7.5)
 7.2045281870976666726e-09L ,   // basis(8)
 4.4229185004672962615e-10L ,   // basis(8.5)
 1.8261499938887221924e-11L ,   // basis(9)
 4.5452628388307088642e-13L ,   // basis(9.5)
 5.7253638111923437033e-15L ,   // basis(10)
 2.70404290721556569e-17L ,   // basis(10.5)
 2.7132171099984410352e-20L ,   // basis(11)
 1.6117375710961183492e-24L ,   // basis(11.5)
 9.6067045396335026575e-32L ,   // basis(12)
 } ; 
const double Poles_24[] = {
-0.82065517417952760226 ,
-0.54920097364808984075 ,
-0.3603190653178175995 ,
-0.22802014939914075353 ,
-0.13618849963046011919 ,
-0.074351497302516889043 ,
-0.035244126673212937406 ,
-0.013246375325256078484 ,
-0.0032976826232791502103 ,
-0.00035807154412069458092 ,
-4.8126755630580574097e-06 ,
-3.5407088073360672255e-12 ,
} ;

const double* const precomputed_poles[] = {
  0, 
  0, 
  Poles_2, 
  Poles_3, 
  Poles_4, 
  Poles_5, 
  Poles_6, 
  Poles_7, 
  Poles_8, 
  Poles_9, 
  Poles_10, 
  Poles_11, 
  Poles_12, 
  Poles_13, 
  Poles_14, 
  Poles_15, 
  Poles_16, 
  Poles_17, 
  Poles_18, 
  Poles_19, 
  Poles_20, 
  Poles_21, 
  Poles_22, 
  Poles_23, 
  Poles_24, 
} ;

const long double* const precomputed_basis_function_values[] = {
  K0, 
  K1, 
  K2, 
  K3, 
  K4, 
  K5, 
  K6, 
  K7, 
  K8, 
  K9, 
  K10, 
  K11, 
  K12, 
  K13, 
  K14, 
  K15, 
  K16, 
  K17, 
  K18, 
  K19, 
  K20, 
  K21, 
  K22, 
  K23, 
  K24, 
} ;

} ; // end of namespace vspline_constants

#define VSPLINE_POLES_H
#endif
