/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform rational b-splines                           */
/*                                                                      */
/*            Copyright 2015, 2016 by Kay F. Jahnke                     */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file prefilter_poles.cc

    \brief calculates the poles of the b-spline prefilter using gsl and BLAS

    this doesn't have to be done for installing vspline if poles.cc is
    already present. Providing degrees up to 24 is just about what gsl
    can handle, with such high degrees the evaluation becomes quite imprecise
    as well, especialy for floats.

    compile with:
    g++ -std=c++11 prefilter_poles.cc -oprefilter_poles -lgsl -lblas
    
    run
    
    ./prefilter_poles > poles.cc

    TODO: could do with some TLC...
*/

#include <iostream>
#include <iomanip>

#include <vigra/array_vector.hxx>
#include <vigra/splines.hxx>
#include <gsl/gsl_poly.h>
#include <vspline/basis.h>

using namespace std ;
using namespace vigra ;

// template < class real_type >
// real_type bspline_basis ( real_type x , int degree , int derivative = 0 )
// {
//   if ( degree == 0 )
//   {
//     if ( derivative == 0 )
//         return x < real_type(0.5) && real_type(-0.5) <= x ?
//                   real_type(1.0)
//                 : real_type(0.0);
//     else
//         return real_type(0.0);
//   }
//   if ( derivative == 0 )
//   {
//     real_type n12 = real_type((degree + 1.0) / 2.0);
//     return (     ( n12 + x )
//                 * bspline_basis ( x + real_type(0.5) , degree - 1 , 0 )
//               +   ( n12 - x )
//                 * bspline_basis ( x - real_type(0.5) , degree - 1 , 0 )
//             )
//             / degree;
//   }
//   else
//   {
//     --derivative;
//     return   bspline_basis ( x + real_type(0.5) , degree - 1 , derivative )
//            - bspline_basis ( x - real_type(0.5) , degree - 1 , derivative ) ;
//   }
// }

template < class T >
ArrayVector<double> 
calculatePrefilterCoefficients(int DEGREE)
{
    ArrayVector<double> res;
    const int r = DEGREE / 2;
    double a[2*r+1] ;
    double z[4*r+2] ;
    cout << "const long double K" << DEGREE << "[] = {" << endl ;
    // we calculate the basis function values at 0.5 intervals
    int imax = 2 * r ;
    if ( DEGREE & 1 )
      imax++ ;
    for(int i = 0; i <= imax ; ++i)
    {
      long double half_i = i / (long double) 2.0 ;
      long double v = vspline::gen_bspline_basis<long double> ( half_i , DEGREE , 0 ) ;
      cout << " " << v << "L ,   // basis(" << half_i << ")" << endl ;
      if ( ! ( i & 1 ) )
      {
        // for even i, we put the value in a[] as well - only even i
        // correspond to the value of the basis function at integral values
        // which we need for the poles
        int ih = i / 2 ;
        a [ r - ih ] = a [ r + ih ] = v ;
      }
    }
    cout << " } ; " << endl ;
        
    if(DEGREE > 1)
    {
        ArrayVector<double> roots;
	
	// we set up the environment gsl needs to find the roots
	gsl_poly_complex_workspace * w 
          = gsl_poly_complex_workspace_alloc (2*r+1);
	// now we call gsl's root finder
        gsl_poly_complex_solve (a, 2*r+1, w, z);
	// and release it's workspace
        gsl_poly_complex_workspace_free (w);

	// we only look at the real parts of the values, which are stored
	// interleaved real/imag. And we take them back to front, even though
	// it doesn't matter to Thevenaz' code which end we start with - but conventionally
	// Pole[0] is the root with the largest absolute, so I stick with that.
        for(int i = 2 * r - 2 ; i >= 0; i-=2)
            if(VIGRA_CSTD::fabs(z[i]) < 1.0)
                res.push_back(z[i]);
    }
    return res;
}

// TODO ugly mishmash of prints and calculations...

void print_poles ( int degree )
{
  ArrayVector<double> res = calculatePrefilterCoefficients<double> ( degree ) ;
  if ( degree > 1 )
  {
    cout << "const double Poles_" << degree << "[] = {" << endl ;
    for ( auto r : res )
      cout << r << " ," << endl ;
    cout << "} ;" << endl ;
  }
}

int main ( int argc , char * argv[] )
{
  cout << setprecision(20) ;
  
  for ( int degree = 0 ; degree < 25 ; degree++ )
    print_poles(degree) ;
  
  cout << noshowpos ;
  cout << "const double* const precomputed_poles[] = {" << endl ;
  cout << "  0, " << endl ;
  cout << "  0, " << endl ;
  for ( int i = 2 ; i < 25 ; i++ )
    cout << "  Poles_" << i << ", " << endl ;
  cout << "} ;" << endl ;
  cout << "const long double* const precomputed_basis_function_values[] = {" << endl ;
  for ( int i = 0 ; i < 25 ; i++ )
    cout << "  K" << i << ", " << endl ;
  cout << "} ;" << endl ;
}
