/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file unary_functor.h

    \brief interface definition for unary functors

    vspline's evaluation and remapping code relies on a unary functor template
    which is used as the base for vspline::evaluator and also constitutes the
    type of object accepted by most of the functions in remap.h.

    This template produces functors which are meant to yield a single output
    for a single input, where both the input and output types may be single
    types or vigra::TinyVectors, and their elementary types may be vectorized.
    The functors are required to provide methods named eval() which are capable
    of performing the required functionality. These eval routines take both
    their input and output by reference - the input is taken by const &, and the
    output as plain &. The result type of the eval routines is void. While
    such unary functors can be hand-coded, the class template 'unary_functor'
    provides services to create such functors in a uniform way, with a specifc
    system of associated types and some convenience code. Using unary_functor
    is meant to facilitate the creation of the unary functors used in vspline.
    
    Using unary_functor generates objects which can be easily combined into
    more complex unary functors, a typical use would be to 'chain' two unary_functors,
    see class template 'chain' below, which also provides an example for the use
    of unary_functor. While this is currently the only explicitly coded combination,
    providing more should be easy and the code would be similar to other expression
    template implementations.
    
    vspline's unary functor objects are composed of two components:
    
    - class template uf_types, to provide a standard set of types
    
    - an 'evaluation policy' providing the actual code for the functor
    
    the policy defines one or several functions eval() handling the operation
    at hand. Using the policy approach seems a roundabout way at first, but
    it makes it quite easy to produce new unary functors. The policy is incorporated
    into the functor by means of a template template argument. The policy shares
    the functor's argument and result type and it's vector width. These template
    arguments are always passed through to the evaluation policy, optionally
    followed by more class type template arguments, which are also passed via
    the functor's template argument list, resulting in a reasonably terse notation.
    Current C++ syntax offers no way to pass on arbitrary variadic template arguments;
    there is a different syntax for type- and non-type arguments. This is annoying
    but can't be helped. In class evaluator (eval.h) I work around this problem by
    creating types from int template arguments using std::numerical_constant.
    
    Class vspline::evaluator is itself coded as a vspline::unary_functor and can
    serve as another example for the use of the code in this file.
*/

#ifndef VSPLINE_UNARY_FUNCTOR_H
#define VSPLINE_UNARY_FUNCTOR_H

#include <vspline/common.h>

namespace vspline {

/// struct uf_types provides the types used throughout vspline in unary functors.
/// These types are pulled in with the macro using_unary_functor_types() below.
/// I keep these type declarations in a separate class because they may be used
/// by other classes (via inheritance) - especially by evaluation policies which
/// have to be explicit about their i/o types (simple policies can avoid refering
/// to the types by using templated eval routines)

// TODO surround the _vsize template argument with #ifdef USE_VC, to make it disappear
// or become 1 in unvectorized code

template < typename IN ,  // argument or input type, like coordinate for interpolators
           typename OUT , // result type, like the type of an interpolation result
           int _vsize >   // vector width. often derived from OUT with vector_traits
struct uf_types
{  
  // number of dimensions. This may well be different for IN and OUT.
  // note how we rely on vigra's ExpandElementResult mechanism to inspect IN and OUT.

  enum { dim_in = vigra::ExpandElementResult < IN > :: size } ;
  enum { dim_out = vigra::ExpandElementResult < OUT > :: size } ;

  // number of elements in simdized data

  enum { vsize = _vsize } ;

  // typedefs for incoming (argument) and outgoing (result) type. These two types
  // are non-vectorized types, like vigra::TinyVector < float , 2 >. Since such types
  // consist of elements of the same type, the corresponding vectorized type can be
  // easily automatically determined.
  
  typedef IN in_type ;
  typedef OUT out_type ;
  
  // elementary types of same. we rely on vigra's ExpandElementResult mechanism
  // to provide these types.
  
  typedef typename vigra::ExpandElementResult < IN > :: type in_ele_type ;
  typedef typename vigra::ExpandElementResult < OUT > :: type out_ele_type ;
  
#ifdef USE_VC
  
  // for vectorized operation, we need a few extra typedefs
  // I use a _v suffix instead of the _type suffix above to indicate vectorized types.

  /// a simdized type of the elementary type of result_type,
  /// which is used for coefficients and results. this is fixed via
  /// the traits class vector_traits (in common.h). Note how we derive
  /// this type using vsize from the template argument, not what
  /// vspline::vector_traits deems appropriate for ele_type - though
  /// both numbers will be the same in most cases.
  
  typedef typename vector_traits < IN , vsize > :: ele_v in_ele_v ;
  typedef typename vector_traits < OUT , vsize > :: ele_v out_ele_v ;
  
  /// vectorized in_type and out_type. with the current implementation this is
  /// a vigra::TinyVector of the ele_v types above, which may only have one member.
  /// one-member TinyVectors are annoying in places - to operate on them, they have
  /// to be 'unpacked' to their single element, or the access has to be done with
  /// a cast, On the other hand, having both single and plural data packaged in the
  /// same container type makes handling them generically easier. Anyway, since the
  /// dimension of incoming and outgoing values is a compile-time constant,
  /// time-critical code can dispatch to specialized code - notably SIMD load/store
  /// operations in perference to SIMD gather/scatters for singular data interacting
  /// with unstrided memory.
  
  typedef typename vector_traits < IN , vsize > :: type in_v ;
  typedef typename vector_traits < OUT , vsize > :: type out_v ;

  /// vsize wide vector of ints, used for gather/scatter indexes
  
  typedef typename vector_traits < int , vsize > :: ele_v ic_v ;

  /// out_type_of provides the result type of an evaluation given it's argument
  /// type. With this declaration, operator() can be expressed.
  
  template < class in_type >
  using out_type_of = typename
    std::conditional
    < std::is_same
      < in_type ,
        in_v
      > :: value ,
      out_v ,
      out_type
    > :: type ;

#else

  /// for non-vectorized code, out_type_of is much simpler:
    
  template < class in_type >
  using out_type_of = out_type ;

#endif

} ;

/// since we want to use the types in class uf_types mulitply, there are
/// macros to pull all the types of class uf_types into a class derived
/// from it. classes which want to use these macros need to invoke them
/// with the base class as their argument.
/// for unverctorized operation we have:

#define using_singular_unary_functor_types(base_type) \
  using typename base_type::in_ele_type ;             \
  using typename base_type::out_ele_type ;            \
  using typename base_type::in_type ;                 \
  using typename base_type::out_type ;                \
  enum { dim_in = base_type::dim_in } ;               \
  enum { dim_out = base_type::dim_out } ;

/// for vectorized operation there are a few more types to pull in:

#define using_simdized_unary_functor_types(base_type) \
  using typename base_type::in_ele_v ;                \
  using typename base_type::out_ele_v ;               \
  using typename base_type::in_v ;                    \
  using typename base_type::out_v ;                   \
  using typename base_type::ic_v ;
  
/// finally a macro automatically pulling in the proper set of type names
/// depending on USE_VC. This is the macro used throughout. Here, vsize is also
/// fixed to the base class' value - or to 1 if USE_VC isn't defined.

#ifdef USE_VC
#define using_unary_functor_types(base_type)    \
  enum { vsize = base_type::vsize } ;           \
  using_singular_unary_functor_types(base_type) \
  using_simdized_unary_functor_types(base_type) 
#else  
#define using_unary_functor_types(base_type)    \
  enum { vsize = 1 } ;                          \
  using_singular_unary_functor_types(base_type) 
#endif

/// struct unary_functor combines the type system above and a specific
/// evaluation policy, which provides the eval() routines needed for a
/// specific operation. Contrary to my initial implementation prescribing
/// an interface by defining two pure virtual functions eval() for
/// unvectorized and vectorized operation, I now rely entirely on the
/// evaluation policy. This gives more flexibility, because now a specific
/// implementation of a unary functor can provide only the eval() variants
/// which will actually be used by calling code, and the policy can also
/// use templates for the evaluation code, which is helpful if both the
/// vectorized and unvectorized code are the same.
///
/// unary_functor inherits from uf_types to have the eval type system,
/// and from the specific evaluation policy to have eval routines.
/// The specification of unary_functor ensures that all components share
/// the same argument type, result type and vector width, since there is only
/// one single point where they are introduced, namely as the first three
/// template arguments in unary_functor's template argument list.
  
// TODO I use the same type system in several places, so it's separate, but it's
// attractive to pull the types right into unary_functor, avoiding the ugly macros.
// The problem is that evluation policies work with types which are isomorphic
// to the types in uf_types, but usually more aptly named: where uf_types
// might have in_type, a policy might have coordinate_type. But the types in
// uf_types are needed to have a handle on their syntactic function, which is
// referred to when unary functors are combined (by incorporation or chaining)

// TODO consider a traits class instead. for vectorized types, we have vector_traits
// already, maybe it would be wise to have something more general like 'eval_traits'

// TODO unary_functor receives trailing template arguments after the policy which
// are passed through to the policy's template instantiation. This is done via
// the variadic template argument 'class ... trailing_args'. I'd like to be able to
// also pass non-type arguments (int, enums...), but afaict the syntax 'typename ...'
// is specific to type arguments, and I don't know of a general syntax allowing
// arbitrary template arguments. This shortcoming limits the usefulness of the
// passing through of the policy's template arguments, but it can be worked around
// by making types out of the noin-type template arguments, like by using
// std::numerical_constant
  
template < typename argument_type ,            // argument and result type are input and
           typename result_type ,              // output of eval()
           int _vsize ,                        // size of vectors to process
           template < typename at ,            // template template argument passing on
                      typename rt ,            // argument_type, result_type and _vsize
                      int vsz ,                // to the evaluation policy
                      typename ... policy_args // plus 0 ot more types to be passed
                    > class eval_policy ,      // to the policy additionally
           typename ... trailing_args >        // these additional types are passed in here
struct unary_functor
// inherit from uf_types to get the standard evaluator type system
: public uf_types < argument_type , result_type , _vsize > ,
  public eval_policy < argument_type ,      // now here's the inheritance from the policy.
                       result_type ,        // it takes argument_type, result_type and
                       _vsize ,             // _vsize plus any trailing args passed to
                       trailing_args ... >  // unary_functor
{
  // pull in the standard evaluator type system. While it would be nice to simply
  // incorporate these typedefs here, I also use them in other classes.
  // inheriting by itself does not provide the types from uf_types, they have to be
  // specifically declared with the using_unary_functor_types() macro above, which
  // has using declarations making the typedefs from uf_types available without
  // explicit reference to class uf_types.
  // an alternative would be to not inherit from uf_types and use typedefs in the
  // macro to introduce the types, effectively using uf_types as a traits class.
    
  typedef uf_types < argument_type , result_type , _vsize > base_type ;
  using_unary_functor_types ( base_type ) ;
  
  // introduce a compact name for the evaluation policy:
  
  using evp = eval_policy < argument_type , result_type , _vsize ,
                            trailing_args ... > ;
                            
  /// variadic constructor template. this passes any constructor
  /// arguments to the evaluation_policy. This looks odd at first, but is necessary:
  /// the evaluation policy may have state which needs to be initialized, and the mere
  /// type of the evaluation policy does not contain information on it's constructor
  /// arguments. But since class unary_functor is a mere shell and does not have
  /// state specific to itself, we can be certain that all arguments passed to it's
  /// constructor are 'meant' for it's evaluation policy, and just pass them on. We
//   /// use perfect forwarding to preserve the arguments as best as we can.
  
  template < class ... ctor_arg_types >
  unary_functor ( ctor_arg_types && ... ctor_args )
  : evp ( std::forward<ctor_arg_types> ( ctor_args ) ... )
  { } ;

  // note how unary_functor contains no eval() routines. These are the domain of
  // the evaluation policy (evp in this case).
  // We rely on evp providing all eval() variants the calling code actually *uses*
  // rather than enforcing a specific interface, like by using pure virtual functions
  // as I did in my initial implementation.
  
  // to be able to declare operator(), we need to derive the return type
  // from the incoming type, using out_type_of from the standard eval type system.
  // alternatively we could use overloading, but then we need separate definitions
  // for all cases and we can't use the generic form here:

  // TODO: clarify what to do when 1D values come into play. currently, in_v
  // and out_v are syntactically nD values with one dimension.
  
  template < typename I ,
             typename O
               = typename base_type::template out_type_of<I>
           >
  O operator() ( const I & i ) const
  {
    O result ;
    evp::eval ( i , result ) ;
    return result ;
  }
} ;

/// class chain is a helper class to easily pass one unary functor's result
/// as argument to another one. We rely on T1 and T2 to provide a few of the
/// standard types used in unary functors. Typically, T1 and T2 will both be
/// vspline::unary_functors, but the type requirements could also be fulfilled
/// manually.
/// This class can also serve as an example for a policy which can be used
/// by a vspline::unary_functor. Such policies have to accept three mandatory
/// template arguments (argument_type, result_type and vsize), plus optionally
/// more types which are specific to the policy. Here, the types of the
/// functors to be chained (T1, T2) are specific to the policy and follow
/// the standard template arguments.

template < typename argument_type ,   // first the three standard template
           typename result_type ,     // arguments
           int vsize ,
           typename T1 ,              // then the two specific arguments
           typename T2 >              // for chain_policy
struct chain_policy
{
  // chaining is only allowed if a set of conditions is fulfilled:
  // T1's input and T2's output type must match argument_type and result_type:
  
  static_assert ( std::is_same < argument_type ,
                                 typename T1::in_type > :: value ,
                  "can only chain unary functors where argument_type == T1::in_type" ) ;

  static_assert ( std::is_same < result_type ,
                                 typename T2::out_type > :: value ,
                  "can only chain unary functors where result_type == T2::out_type" ) ;

  // require a common intermediate type. This requirement is currently omitted
  // - the spec is widened to allow situations where T2::in_type can be constructed from
  // T1::out_type.

//   static_assert ( std::is_same < typename T1::out_type ,
//                                  typename T2::in_type > :: value ,
//                   "can only chain unary functors where T1::out_type == T2::in_type" ) ;

  // we require both functors to share the same vectorization width
                  
  static_assert ( T1::vsize == T2::vsize ,
                  "can only chain unary functors with the same vector width" ) ;

  // hold the two functors as const references

  const T1 & t1 ;
  const T2 & t2 ;
  
  // the constructor initializes these references

  chain_policy ( const T1 & _t1 ,
                 const T2 & _t2 )
  : t1 ( _t1 ) ,
    t2 ( _t2 )
    { } ;

  // the actual eval needs a bit of trickery to determine the type of
  // the intermediate object from the type of the first argument. While
  // it's possible to specialize the code where I == B to omit the
  // intermediate object, I rely on the optimizer to do so, since removing
  // unneccessary intermediates is a trivial task.

  template < typename A ,
             typename B ,
             typename I = typename T1::template out_type_of<A> >
  void eval ( const A & c ,
                    B & result ) const
  {
    I cc ;                    // have an intermediate handy
    t1.eval ( c , cc ) ;      // evaluate first functor into it
    t2.eval ( cc , result ) ; // feed it as input to second functor
  }

} ;

/// struct chain specializes unary_functor with the chain policy above.
/// this is also a handy example for the use of unary_functor, demonstrating how
/// the type handling is reasonably painless using unary_functor. With the
/// alias declaration below, given two unary functors A and B, we can
/// instantiate a 'chain' class as simply as vspline::chain < A , B >.

template < typename T1 ,  // 'chain' objects must be instantiated with two
           typename T2 >  // template arguments, spcifying the functors to use
using chain = typename
 vspline::unary_functor < typename T1::in_type ,  // in_type and out_type are
                          typename T2::out_type , // inferred from T1 and T2
                          T1::vsize ,             // as is the vector width
                          chain_policy ,          // this is the policy argument
                          T1 ,                    // and the trailing types are
                          T2 > ;                  // passed through to the policy

} ; // end of namespace vspline

#endif // VSPLINE_UNARY_FUNCTOR_H

